/*
 * beam.c
 * code to beam texts across the TCP connection following a special protocol
 * source file for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <signal.h>

#include "cancan.h"
#include "term.h"
#include "parse.h"
#include "beam.h"

int gettimeofday();

static int next_digit();
static void edit_message();
static void finish_edit();

editsess *edit_sess;		/* head of session list */

char *edit_start, *edit_end;	/* when starting/leaving editing sessions */

/*
 * read in another decimal. return 0 in case of invalid char.
 */
static int next_digit(val, cp)
unsigned long *val;
char *cp;
{
    unsigned int i;
    return (i = ctoi(*cp)) < 10 ? (*val = *val * 10 + i, 1) : 0;
}

/*
 * Process editing protocol message with len chars.
 * Return number of characters used in the message.
 */
int mp_receive(sesp, data, len)
editsess **sesp;
char *data;
int len;
{
    editsess *s = *sesp;
    char *src = data;
    char mbuf[64];
    int l;
    if(!s) {
	*sesp = s = (editsess*)calloc(1, sizeof(editsess));
	if(!s) syserr("calloc");
#if 0	/* NOTE: a null pointer isn't necessarily all-bits-zero. hmm */
	s->next = s->fname = 0;
#endif
	if(edit_sess) {
	    s->next = edit_sess;
	    s->num = edit_sess->num;
	}
	s->num++;  edit_sess = s;
	src += strlen(MPI);  len -= strlen(MPI);
    }
    if(s->state > 1 && len > (l = s->size - s->pos)) len = l;
    l = len;
    if(l) for(;;) {
	switch(s->state) {
	case 0:					/* 'E' or 'V'	*/
		if(*src != 'E' && *src != 'V') break;
		s->mode = *src;
		src++;  s->state++;  if(!--l) break;
		/* FALLTHROUGH */
	case 1:					/* message size	*/
		while(next_digit(&s->size, src) && (++src, --l));
		if(!l || !s->size) break;
		s->state++;  continue;
	case 4:					/* 'M' if edit	*/
		if(l > s->size) l = s->size;
		len = l;	/* message body starts */
		if(s->mode != 'E') { s->state = 8;  continue; }
		if(*src != 'M') break;
		src++;  s->state++;  if(!--l) break;
		/* FALLTHROUGH */
	case 5:					/* message key	*/
		while(next_digit(&s->key, src) && (++src, --l));
		if(!l || !s->key) break;
		s->state++;  continue;
	case 8:					/* description	*/
		/* keep description length in s->fd; it's not used yet */
		while(isprint(*src)) {
		    if(s->fd < MAXDESCLEN - 1) s->desc[s->fd++] = *src;
		    if(++src, !--l) break;
		}
		if(!l) break;
		fputs(s->desc, stdout);  newline();
		s->state++;
		/* FALLTHROUGH */
	case 2: case 6: case 9:			/* CR's		*/
		s->state++;
		if(*src == '\r' && (++src, !--l)) break;
		/* FALLTHROUGH */
	case 3: case 7: case 10:		/* LF's		*/
		if(*src != '\n') break;
		src++;  --l;  s->state++;
		if(s->state != 11)
		    if(!l) break;
		    else continue;
		if(s->mode) {			/* create file	*/
		    struct timeval tt;
		    sprintf(mbuf, "/tmp/%c%lu", s->mode,
			    s->key ? s->key : (unsigned long)
			    (gettimeofday(&tt, (struct timezone *)0), tt.tv_sec));
		    if((s->fd = open(mbuf, O_WRONLY | O_CREAT, 0666)) < 0) {
			syswarn("#could not create temp file");
			s->mode = 0;
		    } else
			s->fname = str_dup(mbuf);
		}
		if(!l) break;
		/* FALLTHROUGH */
	case 11:				/* write file	*/
		if(!s->mode) {
		    /*
		     * FEATURE: a short packet is assumed to mean
		     * that the data transmission was aborted.
		     */
		    /* if(len < 256 && len < s->size - s->pos) break; */
		    if(len < 256) break;
		} else if(write(s->fd, src, l) < l) {
		    syswarn("#could not write temp file");
		    s->mode = 0;
		}
		src += l;  l = 0;
		if(s->pos + len < s->size) {
		    sprintf(mbuf, "read %lu bytes out of %lu",
			    s->pos + len, s->size);
		    to_footnote(mbuf);
		}
	} /* end switch */
	break;
    } /* end while */
    if(l || (s->state > 4 && (s->pos += len) >= s->size)) {
	*sesp = 0;
	if(l || s->state < 11) {
	    if(s->mode)
		to_footnote("#warning: mume protocol error");
	    finish_edit(s);
	} else {
	    edit_message(s);
	}
    }
    return src - data;
}

/*
 * run the editor or pager, possibly wait for it to finish.
 * NOTE: perhaps we should ignore the environment variables,
 * and use default editor/pager if NOSHELL is defined.
 */
static void edit_message(s)
editsess *s;
{
    char *editor;
    int pid, waitforeditor = 0;
    if(s->mode == 'E') {
	struct stat sbuf;
	if(fstat(s->fd, &sbuf) < 0) syserr("fstat");	/* file is open */
	s->size = sbuf.st_size;
	s->time = sbuf.st_mtime;
	if(!(editor = get_env("CANCANEDITOR")) && !(editor = get_env("EDITOR")))
	    editor = DEFAULT_EDITOR;
    } else if(s->mode == 'V') {
	if(!(editor = get_env("CANCANPAGER")) && !(editor = get_env("PAGER")))
	    editor = DEFAULT_PAGER;
    } else {
	finish_edit(s);
	return;
    }
    close(s->fd);
    /* send an edit_start message (if wanted) */
    if(edit_start && !edit_sess->next)
	parse(edit_start, 0, (struct param *)0);
    if(*editor == '&')
	editor++;
    else {
	waitforeditor = 1;
	reset_terminal();
    }
    flush_output_buffer();
    fflush(stdout);
    s->pid = pid = fork();
    if(!pid) {
	static char *args[] = { "/bin/sh", "-c", insertbuf, 0 };
	sprintf(insertbuf, "%s %s", editor, s->fname);
	set_env("TITLE", s->desc);
	if(!waitforeditor) { close(0);  close(1); }
	execve(args[0], args, environ);
	_exit(1);
    } else if(pid < 0) {
	printf("#fork failed, sorry\n");
	s->mode = 0;
	finish_edit(s);
	if(!waitforeditor) return;
    } else {
	if(!waitforeditor) return;
	attend_child(pid);
	lastline();
    }
    set_terminal();
}

static void finish_edit(s)
editsess *s;
{
    union {
	char buf[BUFSIZE];
	struct stat st;
    } b;
    editsess **sp;
    unsigned long size;
    int i;
    /* MP messages aren't prompted */
    if(s->mode == 'E' && s->fname)
	if((s->fd = open(s->fname, O_RDONLY)) < 0 || fstat(s->fd, &b.st) < 0 ||
		((size = b.st.st_size) == s->size && b.st.st_mtime == s->time))
	    s->mode = 0;
	else {
	    /* file was changed by editor: send back result to server */
#if defined(_POSIX_SOURCE)
	    send_to_host(b.buf, sprintf(b.buf, "%sE%lu\nE%lu\n", MPI,
		    sprintf(b.buf, "_%lu_", s->key) + size, s->key));
#else
	    sprintf(b.buf, "_%lu_", s->key);
	    sprintf(b.buf, "%sE%lu\nE%lu\n", MPI,
		    strlen(b.buf) + size, s->key);
	    send_to_host(b.buf, strlen(b.buf));
#endif
	    for( ; size; size -= i) {
		i = sizeof(b.buf);
		if(i > size) i = size;
		i = read(s->fd, b.buf, i);
		if(i > 0) send_to_host(b.buf, i);
		else { syserr("read temp"); break; }
	    }
	    sprintf(b.buf, "#session %d completed", s->num);
	    to_footnote(b.buf);
	}
    if(s->mode != 'E' && s->key) {		/* cancel editing session */
#if defined(_POSIX_SOURCE)
	send_to_host(b.buf, sprintf(b.buf, "%sE%lu\nC%lu\n", MPI,
		(unsigned long)sprintf(b.buf, "_%lu_", s->key), s->key));
#else
	sprintf(b.buf, "_%lu_", s->key);
	sprintf(b.buf, "%sE%lu\nC%lu\n", MPI,
		(unsigned long)strlen(b.buf), s->key);
	send_to_host(b.buf, strlen(b.buf));
#endif
	sprintf(b.buf, "#session %d cancelled", s->num);
	to_footnote(b.buf);
    }
    if(s->fname) {
	close(s->fd);
	unlink(s->fname);
	free(s->fname);
    }
    for(sp = &edit_sess; *sp; sp = &(*sp)->next)
	if(*sp == s) { *sp = s->next;  break; }
    free((char*)s);
    /* parse (send) the edit_end message if this is the last editor... */
    if(edit_end && !edit_sess)
	parse(edit_end, 0, (struct param *)0);
}

/*
 * cancel an editing session; does not free anything
 * (the child death signal handler will do that)
 */
void cancel_edit(s)
editsess *s;
{
    int sig = s->mode ? SIGTERM : SIGKILL;
    s->mode = 0;
    if(s->pid)	/* Editicide */
	kill(s->pid, sig);
}

/*
 * case child_pid:
 * > 0, do waitpid until the specified process dies, return child status
 * < 0, do waitpid until queue is flushed
 * ==0, do waitpid once, return
 */
int attend_child(child_pid)
int child_pid;
{
    static int status;
    editsess *sp;
    int pid;
    for( ; ; ) {
	pid = waitpid(-1, &status, child_pid > 0 ? 0 : WNOHANG);
	if(pid <= 0) {
	    if(child_pid > 0 && errno != ECHILD) continue;	/* hmm */
	    break;
	}
	for(sp = edit_sess; sp; sp = sp->next) {
	    if(sp->pid != pid) continue;
	    if(status) sp->mode = 0;
	    finish_edit(sp);
	    break;
	}
	if(!child_pid || child_pid == pid) break;
    }
    return status;
}

