/*
 * cancan  --  mud client with telnet protocol
 *
 * written by Mattias Engdegrd (Yorick) (f91-men@nada.kth.se) 1992-95
 * Maintainers: Ilie (finnag@pvv.unit.no) (and Yorick)
 * originally inspired by the Tintin client by Peter Unold but contains no
 * Tintin code.
 * Also greatly improved upon by Vivriel and Thuzzle. As usual, all the
 * developers are in debt to countless users for suggestions and debugging.
 *
 * This program is in the public domain.
 */

#define VERSION "2.6.4 (Yorick, Vivriel, Thuzzle, Ilie 1992-97)"
#define SAVEFILEVER 2

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#ifdef AIX
#include <sys/select.h>
#endif
#include <fcntl.h>
#include <signal.h>
#include <time.h>

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "cmd.h"
#include "beam.h"

int select();
int gettimeofday();

/*	function declarations		*/
static void init_def_file();
static void mainlupe();
static void process_attn();
#ifndef OMIT_SCHEDULE
static struct timeval *check_schedule();
#endif
#ifdef MOTDFILE
static void printmotd();
#endif
static void get_remote_output();
#ifndef OMIT_SESSIONS
static void process_sub_input();
#endif
static void process_input();
static void getuserinput();
static void setsigs();
static void init_env();
static void def_error();
static void read_settings();
static void flashparen();
void termsig();
void pipesig();
void winchsig();
void childsig();

/*
 * A significant overhead could occur in strncmp when variables are
 * heavily used. Thus, we keep a hash of them in a separate array.
 */
static unsigned char hash_var();
static int find_env();

static char *envhash;
static int envsize;
static int envtop;

/*	GLOBALS		*/

char *hostname;
int portnumber;
char *initstr;		/* initial command to run (send) on connect */

#ifndef OMIT_SESSIONS
/* session data indexed by socket descriptor */
static char *sess_id[MAXFDSCAN];	/* session name */
static int sess_active[MAXFDSCAN];	/* active flag */
static int main_sockfd;	/* main session socket fd */
#endif

int sockfd = 1;		/* current socket file descriptor */
fd_set fdset;		/* set of descriptors to select() on */

int linemode;		/* line mode flags (LM_* in cancan.h) */
int attention;		/* special attention flags (ATTN_* in cancan.h) */

/* booleans */
int speedwalk;		/* speedwalk on? */
int wrap = 1;		/* 1 = word wrap active */
int mp_active;		/* 1 = we have #identified ourselves */
int expect_prompt;	/* are we waiting for remote data? */
int new_settings;	/* set this when user defines new alias/action/etc */
int verbose;		/* 1 if command being parsed is user-entered */
int verbatim;		/* 1 = don't expand aliases or process semicolons */

FILE *capfile;		/* capture file or NULL */

/* for line editing */
void (*last_edit_func)();
int edlen;		/* length of current input line */
int pos;		/* cursor position in line */
int flashback;		/* cursor is on excursion, put back there */
char *prefixstr;	/* inserted in the editing buffer each time */
char *footnote;		/* displayed at the bottom (just once) */
char edbuf[LINELEN];	/* line editing buffer */
char promptstr[LINELEN];/* entire prompt string (for redrawing) */
char insertbuf[LINELEN];/* used in many ways. hm */

int main(argc, argv)
int argc; char *argv[];
{
    if(argc < 2 || argc > 3) {
	fprintf(stderr, "usage:\t%s host port\nor\t%s definition-file\n",
		argv[0], argv[0]);
	return 1;
    }
    init_env();		/* set up our environment first */
    settermcap();
    default_bindings();
#ifndef OMIT_PRINTVER
    printver();
#endif
#ifdef MOTDFILE
    printmotd();
#endif
    if(argc == 2) {	/* one arg, assuming it is definition file name */
	init_def_file(argv[1]);
    } else {	/* two args: assume they are hostname and port number */
	hostname = argv[1];
	portnumber = atoi(argv[2]);
	keypad_bindings();
    }
    default_completions();
    set_terminal();
    setsigs();
    mainlupe();
    return 1;
}

/*
 * initialize cancan from definition file or create one
 */
static void init_def_file(df)
char *df;
{
    char buf[LINELEN], *p;
    FILE *f;
    /*
     * rules: if cancan dir is set it is searched first.
     *        if file doesn't exist, it is created there.
     * If a slash appears in the name, the cancan dir isn't used.
     */
    if(!strchr(df, '/') && (p = get_env("CANCANDIR")) && *p) {
	/* BUG: buffer overflow possible. <snicker> */
	sprintf(buf, "%s/%s", p, df);
	if((f = fopen(buf, "r")) || !(f = fopen(df, "r")))
	    df = buf;
    } else
	f = fopen(df, "r");
    set_env("SAVEFILE", df);
    if(f) {
	read_settings(f);
    } else {
	if(!(f = fopen(df, "a+")))
	    exit(syswarn("Could not open definition file"));
	fclose(f);
	printf("Creating %s\nHost name  :", df);
	fflush(stdout);  fgets(buf, sizeof(buf), stdin);
	if(*skipspace(buf) != '\n') {
	    chop_str(buf, '\n');  hostname = str_dup(buf);
	    printf("Port number:");
	    fflush(stdout);  fgets(buf, sizeof(buf), stdin);
	    portnumber = atoi(buf);
	}
	keypad_bindings();
	save_settings();
    }
}

/*
 * show current version
 */
void printver()
{
    printf(
#ifdef __STDC__
	"cancan ver " VERSION "\nCompiled " __DATE__ " (%s %s)\n",
#else
	"cancan ver %s (%s %s)\n", VERSION,
#endif
#ifdef VT100
	"vt100-only",
#else
	"termcaps",
#endif
#ifndef USESGTTY
	"termio"
#else
	"BSD sgtty"
#endif
    );
}

#ifdef MOTDFILE
/*
 * print the message of the day if present
 */
static void printmotd()
{
    char line[LINELEN];
    FILE *f = fopen(MOTDFILE, "r");
    if(f) {
	while(fgets(line, sizeof(line), f))
	    fputs(line, stdout);
	fclose(f);
    }
}
#endif

/*
 * main loop.
 */
static void mainlupe()
{
    int err;
    fd_set readfds;
    struct timeval timeoutbuf, *timeout;

    FD_ZERO(&fdset);
    FD_SET(0, &fdset);		/* standard input (terminal) */
#ifndef OMIT_SESSIONS
    if(hostname && connect_sess((char *)0, initstr, (char *)0, 0) < 0)
	exitcancan("");
#else
    if(!hostname || (sockfd = doconnect(hostname, portnumber)) < 0)
	exitcancan("#please specify correct hostname and port number\n");
    FD_SET(sockfd, &fdset);	/* add the socket to select() set */
    if(initstr) parse(initstr, 0, (struct param *)0);
#endif

    printf("Type #help for help.\n");

    /* make sure the input line is always the bottom one */
    check_winsize();
    lastline();

    /*		main lupe	*/

    for(;;) {

#ifndef OMIT_SESSIONS
	sockfd = main_sockfd;
#endif

#ifndef OMIT_SCHEDULE
	timeout = check_schedule(&timeoutbuf);
#else
	timeout = 0;
#endif

	if(attention) {
	    if(attention & ATTN_FLASHBACK)
#ifndef OMIT_SCHEDULE
		if(!timeout || timeout->tv_sec)
#endif
		{
		    timeout = &timeoutbuf,
		    timeout->tv_sec = FLASHDELAY / 1000000,
		    timeout->tv_usec = FLASHDELAY % 1000000;
		}
	    if(attention & ATTN_FOOTNOTE)
		display_footnote();
	}

	flush_output_buffer();	/* sploooosh */
	fflush(stdout);		/* ditto */

	readfds = fdset;
	/* NOTE: we should pass highest_fd + 1, rather than MAXFDSCAN */
	err = select(MAXFDSCAN, &readfds, (fd_set *)0, (fd_set *)0, timeout);
	if(err < 0 && errno != EINTR) syserr("select");

	if(attention) process_attn();

	if(err > 0) {
#ifndef OMIT_SESSIONS
	    if(FD_ISSET(0, &readfds)) {
		getuserinput();
		if(!--err) continue;
	    }
	    do if(FD_ISSET(sockfd, &readfds)) {
		get_remote_output();
		if(!--err) break;
	    } while(++sockfd < MAXFDSCAN);
#else
	    if(FD_ISSET(0, &readfds))
		getuserinput();
	    if(FD_ISSET(sockfd, &readfds))
		get_remote_output();
#endif
	}
    }
}

static void process_attn()
{
    int attn = attention;
    attention ^= attn;
    if(attn & ATTN_FLASHBACK) {
	moveto(flashback);
	if(!(attn & ~ATTN_FLASHBACK)) return;
    }
    if(attn & ATTN_FOOTNOTE) {
	str_assign(&footnote, (char *)0);
	display_footnote();
	if(!(attn & ~(ATTN_FLASHBACK | ATTN_FOOTNOTE))) return;
    }
    if(attn & ATTN_WINCHSIG) {
	signal(SIGWINCH, winchsig);
	if(check_winsize()) {
	    clear_all();
	    redraw_all();
	}
    }
    if(attn & ATTN_CHILDSIG) {
	/*
	 * NOTE: on some systems (e.g. AIX), setting a sigchld handler
	 * will promptly deliver the signal if waitpid() has anything to
	 * report. Thus, we would get called once more on the next loop,
	 * and do extra signal() and waitpid() calls. Better flush one
	 * child before reinstalling the signal handler.
	 */
	attend_child(0);
	signal(SIGCHLD, childsig);
	attend_child(-1);
    }
}

#ifndef OMIT_SCHEDULE
/*
 * process and update delayed commands list.
 */
static struct timeval *check_schedule(to)
struct timeval *to;
{
    long delta;
    static struct timeval time0;
    if(schedule) {
	delta = -time0.tv_sec;
	gettimeofday(&time0, (struct timezone *)0);
	time0.tv_sec = SCH_HZ * time0.tv_sec +
		time0.tv_usec / (1000000 / SCH_HZ);
	if(delta) delta += time0.tv_sec;
	while(schedule && schedule->delta <= delta)
	    if(schedule->delta >= 0) {
		delta -= schedule->delta;
		schedule->delta = -1;
		if(schedule->next)
		    schedule->next->delta++;
		/*
		 * NOTE: when a command is run while linemode is LM_CHAR,
		 * the output produced (if any) might not look too good since
		 * LF to CRLF mapping is not in effect. This is rare, though.
		 * Besides, in char mode user cannot change the schedule;
		 * perhaps check_schedule should be disabled then.
		 */
		run_command(schedule->commands);
	    } else
		delete_schedulenode(&schedule);
	if(schedule) {
	    schedule->delta -= delta;
	    to->tv_sec = schedule->delta / SCH_HZ;
	    to->tv_usec = (1000000 / SCH_HZ) * (schedule->delta % SCH_HZ);
	    return to;
	}
    }
    time0.tv_sec = 0;
    return 0;
}
#endif

/*
 * get output from the socket and process/display it.
 */
static void get_remote_output()
{
    static char buffer[BUFSIZE];
    int got = read_socket(buffer, sizeof(buffer) - 1);
    if(!got) return;		/* ^ leave room for terminating zero */
#ifdef DEBUGCODE
    /* debug code to see in detail what strange codes come from the server */
    {
	int i;
	for(i = 0; i < got; i++)
	    fprintf(stderr, isprint(buffer[i]) ? "%c" : "[%d]", buffer[i]);
    }
#endif

#ifndef OMIT_SESSIONS
    /* NOTE: main session can be deactivated with #snoop command! */
    if(!sess_active[sockfd]) return;  /* process only active sessions */
#endif
    if(!(linemode & LM_CHAR)) {
	/* line-at-a-time mode: process input in a number of ways */
	clearinpline();
	hilite(0);
#ifndef OMIT_SESSIONS
	if(!is_main_sess()) process_sub_input(buffer, got);
	else
#endif
	process_input(buffer, got);
	redraw_input();
    } else {
	while(write(1, buffer, got) < 0)
	    if(errno != EINTR) { syserr("write"); break; }
    }
}

#ifndef OMIT_SESSIONS
/*
 * process input from remote host (subsidiary session):
 * assert(sizeof(buf) >= size + 1)
 */
static void process_sub_input(buf, size)
char *buf; int size;
{
    char *p;
    int i;
    int back_col0;
    clear_prompt();
    back_col0 = col0; col0 = 0;
    /*
     * FEATURE: we don't keep track of subsession prompt, nor subsession
     * col0 (could use an array for that). So we set col0 to zero when
     * searching actions to trigger. Therefore, '^' anchoring an action
     * makes no difference for subsessions. (Subsession handling really
     * needs a rewrite. At least there's a lot of room for improvement.)
     */
    for(i = 0; i < size; i++) {
	if(buf[i] == '\r') continue;
	for(p = buf; i < size && buf[i] != '\n'; i++)
	    if(buf[i]) *p++ = buf[i];	/* cut out NUL's */
	*p = 0;		/* there is room */
	/*
	 * FEATURE: no wrapping, no marking in subsidiary session
	 * nonprintable chars (escape codes) should be removed too
	 */
	if(search_action(buf, p - buf)) {
	    printf("%.*s",
		    cols_1 - printf("##%s> ", sess_id[sockfd]), buf);
	    newline();
	}
    }
    col0 = back_col0;
    redraw_prompt();
}
#endif

/*
 * process input from remote host:
 * detect special sequences, trigger actions, locate prompt, word-wrap,
 * print to stdout
 * assert(sizeof(buf) >= size + 1)
 */
static void process_input(buf, size)
char *buf; int size;
{
    static editsess *esess;
    char *p;
    int l, i = 0;

    if(expect_prompt) {
	/* NOTE: expect_prompt is zeroed even if we get only a mp message */
	expect_prompt = 0;
	if(esess) esess->mode = 0;	/* sending cancels the session */
    }
    while(i < size) {
	p = buf + i;
	if(esess || (mp_active && !col0 && !strncmp(p, MPI, strlen(MPI))))
	    i += mp_receive(&esess, p, size - i);
	else {
	    /* process block one line at a time: */
	    if(!col0 && *p == '\r') i++;		/* leading CR */
	    for(p = buf; i < size && buf[i] != '\n'; i++)
		if(buf[i]) *p++ = buf[i];		/* cut out NUL's */
	    if(p != buf && *(p - 1) == '\r') p--;	/* trailing CR */
	    *p = 0;
	    if(capfile) fputs(buf, capfile);
	    if(i++ < size) {
		if(capfile) putc('\n', capfile);
		if(!search_action(buf, p - buf))	/* line #gagged */
		    continue;
		if(buf[0])
		    if(wrap) wrap_out(buf);
		    else mark_and_print(buf);
		promptstr[0] = 0;
		col0 = 0;  newline();
	    } else {
		/*
		 * FEATURE: action search on line fragments can give results
		 * differing from a search on a complete line. Alternatively,
		 * the search could be postponed until a newline is read.
		 */
		search_action(buf, p - buf);
		if(!(p = wrap_out(buf))) p = buf;
		else promptstr[0] = 0;
		if(col0 < cols_1 &&
			strlen(p) + (l = strlen(promptstr)) < LINELEN) {
		    strcpy(promptstr + l, p);		/* promptocatenate */
		} else {
		    col0 = 0;  newline();
		}
		break;
	    }
	}
    } /* end while */
}

/*
 * read terminal input and send to parser.
 * decode keys that send escape sequences
 */
static void getuserinput()
{
    static char typed[CAPLEN];	/* chars typed so far (with partial match) */
    static int nchars;		/* the number of them */
    unsigned char c;
    keynode *p;
    /*
     * We have 4 possible line modes:
     * line mode, local echo: line editing functions in effect
     * line more, no echo: sometimes used for passwords, no line editing
     * char mode, no echo: send a character directly, no local processing
     * char mode, local echo: extremely rare, do as above.
     * NOTE: we could try to read more than 1 char at a time.
     */
    if(read(0, &c, 1) != 1)
	syserr("read from stdin");
    if(!(linemode & (LM_CHAR | LM_NOECHO))) {
	/* normal mode (echo, line buffered) */
	if(pager_text && pager(c)) return;
	if(!nchars && isprint(c)) {
	    if(edlen < LINELEN - 1) {
		hilite(1);
		insertchars(&c, 1);
		if(isrparen(c)) flashparen(c);
		last_edit_func = 0;
	   }
	   return;
	}
	/* NOTE: we should ignore a NUL char */
#if !defined(VT100) && !defined(XASCII)
	if(metakey && (c > 0x80)) {		/* metakey set the high bit */
	    c &= 0x7f;
	    if(!nchars) typed[0] = '\033', nchars = 1;
	}
#endif
	/* search function key strings for match */
	typed[nchars++] = c;
	if((p = keydefs))
	    while(strncmp(typed, p->sequence, nchars) && (p = p->next));
	if(p) {
	    if(p->sequence[nchars]) return;	/* partial match */
	    p->funct(p->call_data);
	    last_edit_func = p->funct;
	}
	nchars = 0;
    } else if(!(linemode & LM_CHAR)) {
	/* sending password (line mode, no echo) */
	edbuf[edlen++] = c;
	if(edbuf[0] == '\003') intrsig();
	do {
	    if(c == '\n') newline();
	    else if(edlen < LINELEN) break;
	    send_to_host(edbuf, edlen);
	    expect_prompt++;
	    edlen = 0;
	} while(0);
	edbuf[pos = edlen] = 0;
    } else {
	/* character mode, no line editing */
	while(write(sockfd, &c, 1) < 0)
	    if(errno != EINTR) { syserr("write to socket"); break; }
	if(!(linemode & LM_NOECHO))
	    putchar(c);
    }
}

/*
 * Flash cursor at parentheses that matches c inserted before current pos
 */
static void flashparen(c)
unsigned char c;
{
    int lev = 0, i = pos;
    while(--i >= 0)
	if(isrparen(edbuf[i])) {
	    lev++;
	} else if(islparen(edbuf[i]) && !--lev) {
	    if(LPAREN(c) == edbuf[i]) {
		attention |= ATTN_FLASHBACK;
		flashback = pos;
		moveto(i);
	    }
	    break;
	}
}

#ifdef NEWBIE
/* NOTE: is there *any*one out there using this? should zap it... */
/*
 * Some people just can't write prose. It's horrible, but here is the
 * fix we have been waiting for.
 */
void denewbiefy()
{
  char *p, *q, *r;
  char buf[LINELEN];
  int n;
  int strncasecmp();

  strcpy(buf, edbuf);
  p = buf; r = edbuf;
  while(*p) 
   {
     while(*p && !islabel(*p)) *r++ = *p++;
     q = p;
     while(*p && islabel(*p)) p++;
     if(*p || *q) 
      {
	n = p - q;
	if(strncasecmp(q, "u", n) == 0) 
	 {
	   strcpy(r, "you"); r += 3;
	 }
	else if(strncasecmp(q, "ur", n) == 0) 
	 {
	   strcpy(r, "your"); r += 4;
	 }
	else if(strncasecmp(q, "urs", n) == 0) 
	 {
	   strcpy(r, "yours"); r += 5;
	 }
	else if(strncasecmp(q, "r", n) == 0) 
	 {
	   strcpy(r, "are"); r += 3;
	 }
	else if(strncasecmp(q, "4", n) == 0) 
	 {
	   strcpy(r, "for"); r += 3;
	 }
	else if(strncasecmp(q, "2", n) == 0) 
	 {
	   strcpy(r, "to"); r += 2;
	 }
	else if(strncasecmp(q, "b", n) == 0) 
	 {
	   strcpy(r, "be"); r += 2;
	 } 
	else if(strncasecmp(q, "b4", n) == 0) 
	 {
	   strcpy(r, "before"); r += 6;
	 }
	else 
	 {
	   strcpy(r, q); r += n;
	 }
      }
   }
  *r = '\0';
  edlen = r - edbuf;
}
#endif

/*
 * terminate cancan as cleanly as possible
 */
void exitcancan(message)
char *message;
{
    reset_terminal();
    printf("%s#cancan stopped dancing.\n", message);
    exit(0);
}

#ifndef NOSHELL
/*
 * execute a shell command and return the result
 */
int execute(arg)
char *arg;
{
    int pid;
    flush_output_buffer();
    fflush(stdout);
    pid = fork();
    if(!pid) {
	static char *args[] = { "/bin/sh", "-c", 0, 0 };
	args[2] = arg;
	execve(args[0], args, environ);
	_exit(1);
    } else if(pid < 0) {
	printf("#fork failed, sorry\n");
    } else {
	pid = attend_child(pid);
    }
    return pid;		/* return the process exit status */
}
#endif

/*
 * set up our signal handlers
 */
static void setsigs()
{
    signal(SIGCHLD, childsig);
    signal(SIGTERM, termsig);
    signal(SIGPIPE, pipesig);
    signal(SIGWINCH, winchsig);
}

void winchsig()
{
    attention |= ATTN_WINCHSIG;
}

void childsig()
{
    attention |= ATTN_CHILDSIG;
}

void pipesig()
{
    exitcancan("\n#broken pipe.\n");
}

void termsig()
{
    exitcancan("\n#termination signal.\n");
}

void intrsig()
{
    if(sockfd > 1) close(sockfd);
    exitcancan("\n#interrupted.\n");
}

/*
 * suspend ourselves
 */
void suspend_cancan()
{
    if(signal(SIGTSTP, SIG_IGN) == SIG_DFL) {
	/* NOTE: we should skip over the edit line here! */
	reset_terminal();  fflush(stdout);
	signal(SIGTERM, SIG_IGN);
	signal(SIGTSTP, SIG_DFL);
	kill(getpid(), SIGTSTP);
	set_terminal();
	signal(SIGTERM, termsig);
	check_winsize();	/* in case size changed meanwhile */
	lastline();		/* so we know where the cursor is */
	redraw_all();
    } else
	to_footnote("#your shell probably lacks job control.");
}

/*
 * print an error message
 */
int syswarn(msg)
char *msg;
{
    if(errno > 0 && errno < sys_nerr)
	printf("%s (%s)\n", msg, sys_errlist[errno]);
    else
	printf("%s\n", msg);
    return -1;
}

/*
 * print system call error message and terminate
 */
void syserr(msg)
char *msg;
{
    /* NOTE: we can't be called before set_terminal(). hmm */
    reset_terminal();
    fprintf(stderr, "\n#fatal system call error: %s (", msg);
    if(errno > 0 && errno < sys_nerr)
	fprintf(stderr, "%s)\n", sys_errlist[errno]);
    else
	fprintf(stderr, "%d)\n", errno);
    exit(1);
}

static void def_error(ln)
int ln;
{
    static int n_err;
    printf("bad definition file format, line %d\n", ln);
    if(++n_err >= 10) exit(1);
}

/*
 * read definitions from file
 */
static void read_settings(f)
FILE *f;
{
    char buf[LINELEN];
    char *p, *cmd;
    int ln;

    for(ln = 1; fgets(buf, sizeof(buf), f); ln++) {
	if(*buf == '\n') continue;
	p = chop_str(cmd = buf, ' ');
	if(p && *cmd == '#') {
	    chop_str(p, '\n');	/* strip lf */
	    if(!strcmp(++cmd, "action"))
		parse_action(p);
	    else if(!strcmp(cmd, "alias"))
		parse_alias(p);
	    else if(!strcmp(cmd, "bind"))
		parse_bind(p);
	    else if(!strcmp(cmd, "mark"))
		parse_marker(p);
	    else if(!strcmp(cmd, "init"))
		str_assign(&initstr, p);
#if 0
	    else if(!strcmp(cmd, "hilite"))
		attr_string(edattr = parse_attributes(p), edattrbeg, edattrend);
#endif
	    else if(!strcmp(cmd, "savefile")) {
		if(atoi(p) > SAVEFILEVER)
		    printf("Warning: this cancan version is too old!\n");
	    } else if(!strcmp(cmd, "host")) {
		char *rest = chop_str(p, ' ');
		if(rest && (portnumber = atoi(rest)))
		    str_assign(&hostname, p);
		else
		    def_error(ln);
	    } else
		def_error(ln);
	} else
	    def_error(ln);
    }
    fclose(f);
}

/*
 * save settings in definition file
 */
void save_settings()
{
    char buf[LINELEN], *cp;
    FILE *f;
    aliasnode *al;
    actionnode *ac;
    marknode *m;
    keynode *k;

    if(!(cp = get_env("SAVEFILE"))) return;
    if(!(f = fopen(cp, "w"))) {
	printf("#error: can't write to definition file %s\n", cp);
	return;
    }
    fprintf(f, "#savefile %d\n", SAVEFILEVER);
    if(hostname)
	fprintf(f, "#host %s %d\n", hostname, portnumber);
    if(initstr)
	fprintf(f, "#init %s\n", initstr);
    /* NOTE: #hilite, #speedwalk, #wrap states should all be saved */
#if 0
    if(edattr != NOATTRCODE)
	fprintf(f, "#hilite %s\n", attr_name(edattr));
#endif
    for(al = aliases; al; al = al->next)
	fprintf(f, "#alias %s=%s\n", al->name, al->subst);
    for(ac = actions; ac; ac = ac->next) {
	escape_specials(ac->pattern, buf, "\\\"");
	fprintf(f, "#action %c%s \"%s\"=%s\n", (ac->active) ? '+' : '-',
		ac->label, buf, ac->commands);
    }
    for(m = markers; m; m = m->next) {
	escape_specials(m->pattern, buf, "\\= ");
	fprintf(f, "#mark %s=%s\n", buf, attr_name(m->attrcode));
    }
    /* NOTE: deleted default keys should be saved too, as "#bind <key>=" */
    for(k = keydefs; k; k = k->next)
	if(k->call_data) {
	    escape_seq(k->sequence, buf);
	    fprintf(f, "#bind %s %s=%s\n", k->name, buf, k->call_data);
	}
    fclose(f);
}

/*
 * if alias recording is active, record given line in that alias
 */
void record_line(line)
char *line;
{
    static int i;
    aliasnode *np;
    if(!(np = *lookup_alias(recording_name))) {
	i = 0;
	add_aliasnode(recording_name, line);
    } else {
	/*
	 * FEATURE: if both alias name and the line to be recorded
	 * are very long, total alias lenght may grow too large.
	 */
	char new_line[LINELEN];
	int room = LINELEN - 10 - strlen(line);
#if defined(_POSIX_SOURCE)
	int len = sprintf(new_line, "%s", np->subst);
#else
	int len = strlen(np->subst);
	strcpy(new_line, np->subst);
#endif
	if(len >= room - LINELEN / 2) {	/* leave ample room for editing */
	    do {
#if defined(_POSIX_SOURCE)
		len = sprintf(new_line, "%s%d", recording_name, ++i);
#else
		sprintf(new_line, "%s%d", recording_name, ++i);
		len = strlen(new_line);
#endif
	    } while(*lookup_alias(new_line));
	    if(len >= room) return;
	    add_aliasnode(new_line, np->subst);
	}
	sprintf(new_line + len, ";%s", line);
	str_assign(&np->subst, new_line);
    }
}

void to_footnote(msg)
char *msg;
{
    attention |= ATTN_FOOTNOTE;
    str_assign(&footnote, msg);
}

char *str_dup(s)
char *s;
{
    char *k = (char *)malloc(strlen(s) + 1);
    if(!k) syserr("malloc");
    return strcpy(k, s);
}

void str_assign(str, s)
char **str, *s;
{
    /* NOTE: old string is freed before allocating the new one */
    if(*str) free(*str);
    *str = s ? str_dup(s) : 0;
}

/*
 * return the hash value of a string
 */
static unsigned char hash_var(str)
char *str;
{
    char *s = str;
    char c, z = 0, x = 0;
    while((c = *s) && c != '=')
	s++, x ^= z += c;
    return x;
}

/*
 * initialize our variables
 */
static void init_env()
{
    char **new_env, *new_hash;
    char **oldenv = environ;
    int i = 0;
    while(oldenv[i++]);
    new_env = (char **)malloc(i * sizeof(char *));
    new_hash = (char *)malloc(i);
    /* we are called before set_terminal -- can't syserr */
    if(!new_env || !new_hash)
	exit(syswarn("Unable to allocate own environment"));
    envtop = --i;
    new_env[i] = 0;
    while(--i >= 0)
	new_hash[i] = hash_var(new_env[i] = str_dup(oldenv[i]));
    environ = new_env;
    envhash = new_hash;
}

/*
 * look up a variable in the environ array
 * return the index to it, or -1 if not found
 */
static int find_env(var, len)
char *var;
int len;
{
    int i = envtop;
    char *s, c = hash_var(var);
    while(--i >= 0) {
	while(envhash[i] != c && --i >= 0);
	if(i < 0) break;
	if(strncmp(s = environ[i], var, len)) continue;
	if(!s[len] || s[len] == '=') break;
    }
    return i;
}

/*
 * analogous to getenv, but with fast hash lookup
 */
char *get_env(variable)
char *variable;
{
    char *s = variable;
    int len = strlen(variable);
    int i = find_env(s, len);
    if(i < 0) return 0;
    s = environ[i] + len;
    return *s ? s + 1 : s;
}

/*
 * set or unset (when value == NULL) an environment variable
 */
void set_env(variable, value)
char *variable, *value;
{
    int i = find_env(variable, strlen(variable));
    if(value) {
	char buf[LINELEN];
	if(i < 0) {
	    i = ++envtop;
	    if(i >= envsize) {
		int new_size = (i | 63) + 1;
		envsize = new_size;
		environ = (char **)realloc(environ, new_size * sizeof(char *));
		if(!environ || !(envhash = (char *)realloc(envhash, new_size)))
		    syserr("realloc");
	    }
	    /* the env array needs to be zero terminated when we execve() */
	    environ[i] = 0;
	    envhash[--i] = hash_var(variable);
	}
	sprintf(buf, "%s=%s", variable, value);
	str_assign(environ + i, buf);
    } else if(i >= 0) {
	int k;
	free(environ[i]);
	k = --envtop;
	if(k != i) {
	    envhash[i] = envhash[k];
	    environ[i] = environ[k];
	}
	environ[k] = 0;
    }
}

#ifndef OMIT_SESSIONS
/*
 * Below are multiple-session support functions.
 *
 * NOTE: sess_active[0] holds total no. of connections
 */

/*
 * return TRUE if current session being processed is main connection
 */
int is_main_sess()
{
    return (sockfd == main_sockfd);
}

/*
 * return session's index or socket descriptor given id,
 * or 0 if invalid id is given
 */
int find_sess(id, notify)
char *id;
int notify;
{
    int i, n;
    for(i = main_sockfd, n = sess_active[0]; n; i++)
	if(sess_id[i]) {
	    n--;
	    if(!strcmp(sess_id[i], id))
		return i;
	}
    if(notify)
	printf("#no such connection\n");
    return 0;
}

/*
 * connect another session
 */
int connect_sess(id, init, host, port)
char *id;	/* session id - NULL if main session */
char *init;	/* optional init string */
char *host;
int port;
{
    int fd;
    if((!host && !(host = hostname)) || (!port && !(port = portnumber))) {
	printf("#please specify both hostname and port number.\n");
	return -1;
    }
    /* dial the number by moving the right index in small circles */
    if((fd = doconnect(host, port)) < 0)
	return fd;
    FD_SET(fd, &fdset);		/* add socket to select() set */
    sess_active[0]++;
    if(!main_sockfd)		/* save main socket fd */
	main_sockfd = sockfd = fd;
    if(!id) {			/* main session */
	sess_active[fd] = 1;
	sess_id[fd] = "main";
    } else {
	sess_active[fd] = !init;
	sess_id[fd] = str_dup(id);
    }
    if(init) parse(init, 0, (struct param *)0);
    return fd;
}

/*
 * disconnect session
 */
void disconnect_sess(sfd)
int sfd;
{
    close(sfd);
    printf("#connection \"%s\" closed.\n", sess_id[sfd]);
    if(sfd == main_sockfd)		/* main connection closed */
	exitcancan("");
    FD_CLR(sfd, &fdset);
    sess_active[0]--;
    sess_active[sfd] = 0;
    str_assign(sess_id + sfd, (char *)0);
}

/*
 * toggle output display from another session
 */
void snoop_sess(id)
char *id;
{
    int sfd = find_sess(id, 1);
    if(sfd)
	sess_active[sfd] ^= 1;
}

/*
 * show open connections
 */
void show_connections()
{
    int i, n = sess_active[0];
    if(n) {
	sprintf(insertbuf, "#The following connection%s open:",
		(n == 1) ? " is" : "s are");
	for(i = main_sockfd; to_pager(insertbuf), --n >= 0; i++) {
	    while(!sess_id[i]) i++;
	    sprintf(insertbuf, "##%-8s%s", sess_id[i],
		    !sess_active[i] ? "" :
		    i == main_sockfd ?  " (active)" : " (snooped)");
	}
    } else
	printf("#No open connections.\n");
}
#endif
