/*
 * parse.c
 * command parsing, argument parsing routines for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "cmd.h"

FILE *popen();
int pclose();
long strtol();

static void subst_param();
#ifdef XEXPR
static char *get_tok();
static char *brace_var();
#endif
static char *getval();
static char *calc();
static struct param *match_action();
static int exec_action();
static void split_arg();
static char *skip_arg();

#ifdef OWN_CTYPE
const unsigned char ct_array[256] = {
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    64, 1, 1, 1, 1, 1, 1, 1,33,17, 1, 1, 1, 1, 1, 1,
     5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 1, 1, 1, 1, 1, 1,
     1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,33, 1,17, 1, 9,
     1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
     3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,33, 1,17, 1, 0,
#if defined(XASCII)
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
#endif
};
#endif

/*
 * silent = 1 if the line should not be echoed, 0 otherwise.
 * if invoked by alias or action, params are passed along.
 */
void parse(cmdline, silent, param)
char cmdline[LINELEN];
int silent;
struct param param[10];
{
    static defnode *recursed_aliases;
    char cmd[LINELEN], buf[LINELEN];
    char *s, *d;
    int i, subst;
    aliasnode *np;

    if(!brace_balance(cmdline)) return;

    s = skipspace(cmdline);			/* line enclosed in braces? */
    if(*s == '{' && !*skipspace((d = skip_arg(++s, '}')) + 1))
	memcpy(cmd, s, d - s), cmd[d - s] = 0, subst = 1;
    else
	strcpy(cmd, cmdline), subst = 0;

    /* PROCESS THE COMMANDS ONE BY ONE */
    s = cmd;
    do {
	cmdline = chop_arg(s, CMDSEP);		/* chop into commands */
	if(recording_name && *s != '#' && !param) record_line(s);

	buf[0] = buf[sizeof(buf) - 1] = 0;

	if((d = skip_arg(s, ' '))) *d = 0;	/* pick out the first word */

	if((np = *lookup_alias(s)) && (*lookup(s, &recursed_aliases) ||
		(recording_name && !strcmp(s, recording_name)))) np = 0;

	if(np) { if(d) *d = ' '; }
	else if(*s == '#') { if(d) d = skipspace(d + 1); }
	else { if(d) *d = ' '; d = s; }		/* restore line */

	if(d)
	    if(subst) subst_param(buf, d, param);
	    else strcpy(buf, d);

	if(param && !cmdline && !subst && recursed_aliases && param[0].len) {
	    /* append arguments */
	    int n = sizeof(buf) - (i = strlen(buf));
	    if(param[0].len < n) n = param[0].len + 1;
	    memcpy(buf + i, param[0].s, n);
	}
	if(buf[sizeof(buf) - 1]) {
	    buf[sizeof(buf) - 1] = 0;
	    printf("#warning: line truncated\n");
	}

	if(np) {				/* process an alias */
	    struct param new_param[10];
	    defnode ban_alias;
	    ban_alias.next = recursed_aliases;  ban_alias.field = s;
	    if(d) *d = 0;			/* cut out first word again */
	    split_arg(buf, new_param);
	    recursed_aliases = &ban_alias;	/* block alias recursion */
	    parse(np->subst, silent, new_param);
	    recursed_aliases = ban_alias.next;
	} else if(*s != '#') {			/* text to send		*/
	    /* NOTE: should we echo the walk sequence also if !silent? */
	    if(buf[0] && (!buf[1] || speedwalk) && parse_walk(buf)) continue;
	    if(!subst) unescape(buf);
	    if(!silent) printf("[%s%s%s]\n", edattrbeg, buf, edattrend);
	    buf[i = strlen(buf)] = '\n';
	    send_to_host(buf, i + 1);
	    expect_prompt++;
	} else if(*++s) {			/* "# " is a comment	*/
	    if(isdigit(*s) && !*skipnum(s)) {
		if((i = atoi(s)) > 0)
		    while(i--) parse(buf, 1, (struct param *)0);
		else
		    printf("#bogus repeat count\n");
	    } else
#ifndef OMIT_SESSIONS
	    if(*s == '#') {			/* a session is specified */
		int backfd = sockfd;
		int back_expect_prompt = expect_prompt;
		if((sockfd = find_sess(s + 1, 1))) {
		    parse(buf, 1, (struct param *)0);
		    if(!is_main_sess())
			expect_prompt = back_expect_prompt;
		}
		sockfd = backfd;
	    } else if(!is_main_sess())
		printf("#command not allowed on subsidiary connection\n");
	    else
#endif
#ifndef NOSHELL
	    if(*s == '?' && !*(s + 1)) {	/* #? command	*/
		i = execute(buf);
		if(!(s = cmdline)) break;
		cmdline = chop_arg(s, CMDSEP);
		if(!i) {
		    parse(s, silent, param);
		    return;
		}
	    } else
#endif
	    /*
	     * NOTE: perhaps we should keep a parse_context structure, or
	     * similar, to pass on to commands, so that #add, #? could be
	     * moved to cmd.c. Or maybe let the commands do their own args
	     * expansions a la pull_arg(). Then the double unescaping with
	     * #echo command could be avoided, etc., etc. <puke>
	     */
	    if(param && !recursed_aliases && !strcmp(s, "add")) {
		if(param[0].len) {
		    sprintf(buf, "%.*s", param[0].len, param[0].s);
		    put_word(buf);
		}
	    } else
	    for(i = 0; ; i++)
		if(!commands[i].name) {
		    printf("#unknown cancan command '%s'\n", s);
		    break;
		} else if(commands[i].funct && !strcmp(s, commands[i].name)) {
		    (*commands[i].funct)(buf);
		    if(line0 < lines_1) lastline();
		    break;
		}
	}
    } while((s = cmdline));
}

/*
 * execute walk if word is valid [speed]walk sequence -
 * return 1 if walked, 0 if not
 */
int parse_walk(word)
char *word;
{
    static char wcmd[2] = { ' ', '\n' };
    int n = strlen(word);

    if(!n || !strchr("nsewud", word[n - 1]) ||
	    strspn(word, "nsewud0123456789") != n)
	return 0;

    for( ; *word; word++) {
	n = 1;
	if(isdigit(*word) && !(n = strtol(word, &word, 10))) continue;
	expect_prompt++;
	wcmd[0] = *word;
	do {
	    send_to_host(wcmd, 2);
	    if(is_main_sess())		/* yargh */
		add_dir_to_map(*word);
	} while(--n);
    }
    return 1;
}

/*
 * verify that the braces in the line are properly balanced
 */
int brace_balance(str)
char *str;
{
    char *message = "#unbalanced %c near \"%.16s\"\n";
    char *d, *s = str;
    if(skip_arg(s, '\0')) return 1;
    if((d = skip_arg(s, '}')))
	printf(message, *d, (d - str > 15) ? d - 15 : str);
    else {
	do {
	    d = skip_arg(s, '{');
	    s = skip_arg(d + 1, '}');
	} while(s++);
	printf(message, *d, d);
    }
    return 0;
}

#if 0
/*
 * verify the label and command
 * for use with #action #alias #bind #identify #init #record #schedule...
 */
int valid_label_cmd(msg, lbl, cmd)
char *msg, *lbl, *cmd;
{
    if(lbl && skip_arg(lbl, ' '))
	printf("#%s label should not contain a space.\n", msg);
    else if(cmd && !skip_arg(cmd, '\0'))
	printf("#%s: unbalanced braces in command.\n", msg);
    else
	return 1;
    return 0;
}
#endif

/*
 * {} *must* be balanced in input string. input will be trashed.
 * leave {...} contents intact, otherwise unescape and replace
 * &0 .. &9, &label, &{label}, &[expression], &(command).
 */
static void subst_param(buf, str, param)
char *buf, *str;
struct param param[10];
{
    char *src, *s = str;
    int i, room = LINELEN;

    while(*s) {
	if(*s != '&') {
	    if(*s != '{') {
		if(*s == '\\' && !*++s) break;
		if(room) --room, *buf++ = *s++;
		continue;
	    } else {
		src = s;  s = skip_arg(s + 1, '}') + 1;
		i = s - src;
	    }
	} else switch(*++s) {
	case '&':				/* delayed &	*/
	    src = s;  for(i = 1; *++s == '&'; i++);
	    break;
	case '{':				/* &{variable}	*/
	    s = skip_arg((src = s + 1), '}');
#ifndef XEXPR
	    {		/* {} must be balanced, no check here	*/
		char sbuf[LINELEN];
		*s++ = 0;  subst_param(sbuf, src, param);
		src = get_env(sbuf);
	    }
#else
	    src = brace_var(src);
#endif
	    if(!src) continue;
	    i = strlen(src);
	    break;
	case '[':				/* &[expr]	*/
	    s = skip_arg((src = s + 1), ']');
	    if(*s) {
		static char vbuf[24];  long val;
#ifndef XEXPR
		char sbuf[LINELEN];
		*s++ = 0;  subst_param(sbuf, src, param);
		if(!(src = calc(sbuf, &val)) || *src) continue;
#else
		*s++ = 0;
		if(!(src = calc(src, &val)) || *src) continue;
#endif
#if defined(_POSIX_SOURCE)
		i = sprintf(src = vbuf, "%ld", val);
#else
		sprintf(src = vbuf, "%ld", val);
		i = strlen(src);
#endif
		break;
	    } else
		printf("#unbalanced [ near \"&[%.16s\"\n", src);
	    continue;
	case '(':				/* &(command)	*/
	    s = skip_arg((src = s + 1), ')');
	    if(*s) {
		FILE *fp;
		*s++ = 0;
		unescape(src);
		/* NOTE: some systems may need a fflush() before popen()? */
		if((fp = popen(src, "r"))) {
		    if((i = fread(buf, 1, room, fp)) > 0) {
			if(buf[i - 1] == '\n') i--;
			buf += i;  room -= i;
		    }
		    pclose(fp);
		}
	    } else
		printf("#unbalanced ( near \"&(%.16s\"\n", src);
	    continue;
	default:				/* &N or &label	*/
	    if(isdigit(*s)) {
		i = ctoi(*s++);
		if(!param || !param[i].len) continue;
		src = param[i].s;  i = param[i].len;
	    } else {
		for(src = s; islabel(*s); s++);
		i = *s;  *s = 0;
		src = get_env(src);  *s = i;
		if(!src) continue;
		i = strlen(src);
	    }
	} /* switch ends */
	if(room < i) i = room;
	memcpy(buf, src, i);
	buf += i;  room -= i;
    } /* end while */
    if(room) *buf = 0;
}

#ifdef XEXPR

#define T_NOP	0
#define T_MUL	1
#define	T_DIV	2
#define T_MOD	3
#define T_ADD	4
#define T_SUB	5
#define T_LT	6
#define T_GT	7
#define T_AND	8
#define T_OR	9
#define T_XOR	10
#define T_EQ	11
#define T_NE	12
#define T_SHL	13
#define T_SHR	14
#define T_LE	15
#define T_GE	16
#define T_LAND	17
#define T_LOR	18
#define N_PR	11		/* number of precedence levels */
static char precedence[] = { 0,10,10,10,9,9,7,7,5,3,4,6,6,8,8,7,7,2,1 };
static char *opstring = "*/%+-<>&|^=!";

static char *get_tok(str, op)
char *str, *op;
{
    char *s;
    int c = 0;
    if(*str && (s = strchr(opstring, *str))) {
	c = s - opstring + 1;  str++;
	switch(c) {
	    case T_EQ:
	    case T_NE:
		if(*str == '=') str++;
		else c = T_NOP;
		break;
	    case T_LT:
		if(*str == '<') { c = T_SHL;  str++; }
		else if(*str == '=') { c += T_LE - T_LT;  str++; }
		break;
	    case T_GT:
		if(*str == '>') { c = T_SHR;  str++; }
		else if(*str == '=') { c += T_GE - T_GT;  str++; }
		break;
	    case T_AND:
		if(*str == '&') { c += T_LAND - T_AND;  str++; }
		break;
	    case T_OR:
		if(*str == '|') { c += T_LOR - T_OR;  str++; }
		break;
	}
    }
    *op = c;
    return str;
}

static char *brace_var(str)
char *str;
{
    char buf[LINELEN];
    buf[sizeof(buf) - 1] = 0;
    subst_param(buf, str, (struct param *)0);
    return buf[sizeof(buf) - 1] ? 0 : get_env(buf);
}
#endif

/*
 * read a value from string; recurse into calc() if parentheses are found.
 * return string pointer properly advanced, NULL signifies failure.
 */
static char *getval(str, val)
char *str; long *val;
{
#ifndef XEXPR
    char *s = str;
    if(*s == '(') {
	s = calc(s + 1, val);
	return (s && *s++ == ')') ? s : 0;
    }
    *val = strtol(s, &str, 10);
    return s != str ? str : 0;
#else
    char *s = str, *k = 0;
    if(*s == '(') {
	s = calc(s + 1, val);
	return (s && *s++ == ')') ? s : 0;
    } else if(!isdigit(*s)) {
	if(*s == '&') s++;
	if(*s == '{') {
	    k = skip_arg(++s, '}');  *k++ = 0;
	    s = brace_var(s);
	    if(!s) return 0;
	} else {
	    char c;
	    k = s;  while(islabel(*k)) k++;
	    if(k == s) return 0;
	    c = *k; *k = 0; s = get_env(s); *k = c;
	}
    }
    *val = strtol(s, &str, 10);
    return s == str ? 0 : k ? k : str;
#endif
}

/*
 * calculate the value of a simple expression. () + - * / % are recognized.
 * return advanced string pointer, or NULL in case of syntax error.
 */
static char *calc(str, val)
char *str; long *val;
{
#ifndef XEXPR
    long v1, v2;
    char op, sign = 0;
    char *s = str;
    if(*s == '-') sign = 1, s++;
    for(*val = 0; (s = getval(s, &v1)); s++) {
	while(op = *s, op == '*' || op == '/' || op == '%')
	    if(!(s = getval(s + 1, &v2))) return 0;
	    else if(op == '*') v1 *= v2;
	    else if(!v2) return 0;
	    else if(op == '/') v1 /= v2;
	    else v1 %= v2;
	*val += sign ? -v1 : v1;
	if(op == '-') sign = 1;
	else if(sign = 0, op != '+') return s;
    }
    return 0;
#else /* XEXPR */
    long v[N_PR], vn1;
    char op[N_PR];
    char *u_s, *s = str;
    int u_ops, o1, o2, n = 0;
    do {
	for(u_s = s; *s == '+' || *s == '-' || *s == '~' || *s == '!'; s++);
	u_ops = s - u_s;
	s = getval(s, v + n);
	if(!s) break;
	/* HANDLE UNARY OPERATORS */
	while(u_ops)
	    if(u_s[--u_ops] == '-') v[n] = -v[n];
	    else if(u_s[u_ops] == '~') v[n] = ~v[n];
	    else if(u_s[u_ops] == '!') v[n] = !v[n];
	s = get_tok(s, op + n);
	/* HANDLE BINARY OPERATORS */
	while(n && precedence[o2 = op[n]] <= precedence[o1 = op[n - 1]])
	    switch(op[--n] = o2, vn1 = v[n + 1], o1) {
		case T_EQ: v[n] = v[n] == vn1;  break;
		case T_NE: v[n] = v[n] != vn1;  break;
		case T_LE: v[n] = v[n] <= vn1;  break;
		case T_GE: v[n] = v[n] >= vn1;  break;
		case T_LT: v[n] = v[n] < vn1;  break;
		case T_GT: v[n] = v[n] > vn1;  break;
		case T_LOR: v[n] = v[n] || vn1;  break;
		case T_LAND: v[n] = v[n] && vn1;  break;
		case T_SHL: v[n] <<= vn1;  break;
		case T_SHR: v[n] >>= vn1;  break;
		case T_OR: v[n] |= vn1;  break;
		case T_XOR: v[n] ^= vn1;  break;
		case T_AND: v[n] &= vn1;  break;
		case T_ADD: v[n] += vn1;  break;
		case T_SUB: v[n] -= vn1;  break;
		case T_MUL: v[n] *= vn1;  break;
		case T_DIV: v[n] /= vn1;  break;
		case T_MOD: v[n] %= vn1;  break;
	    }
    } while(op[n++]);
    *val = v[0];
    return s;
#endif
}

/*
 * compute the minimum match length of a pattern
 * (length of shortest string that can match)
 */
int min_match_len(pat)
char *pat;
{
    int n;
    if(*pat == '^') pat++;
    for(n = 0; *pat; pat++)
	if(*pat != '&') n++;
	else if((unsigned)ctoi(pat[1]) < 10) pat++;
    return n;
}

/*
 * match action containing &0..&9 and return actual params in param array
 *
 * NOTE: if pattern contains a sequence of &n-s, all but last are ignored.
 * FEATURE: & should not match an empty word.
 * FEATURE: we should try to find the longest possible match.
 * FEATURE: can't match '&' as plain text in pattern (its always special).
 * FEATURE: no backtracking used, therefore a match with & might fail, even
 * though respective regexp would not, e.g. pat:"a & b", src:"a a a b" fails.
 */
static struct param *match_action(pat, src)
char *pat, *src;
{
    static struct param param[10];
    static int dirtyregs;
    char *mpat = insertbuf;
    char *s = src, *p = pat;
    int mregs, prm;
    char f, mflag = 1;	/* 1 - at pattern beginning, 2 - single word match */
    if(*p == '^') {			/* anchor match at line start */
	register int ofs = s - p - 1;
	if(col0) return 0;
	while(*++p && *p != '&')
	    if(*p != p[ofs]) return 0;
	s = p + ofs;
	mflag = 0;
    }
    mregs = 0;  prm = -1;  param[0].s = src;
    while(*p) {
	if(*p == '&') do {		/* pick out parameters */
	    p++;  prm = ctoi(*p++);
	    if((unsigned)prm >= 10)	/* support single & for back compat */
		p--, prm = 0, mflag |= 2;	/* match a single word only */
	} while(*p == '&');
	if(*p) {			/* attempt match */
	    register char *mp = mpat;
	    do *mp++ = *p++; while(*p && *p != '&');
	    *mp = '\0';
	    if(prm >= 0)
		param[prm].s = s;
	    s = strstr(s, mpat);
	    if(!s) {
		dirtyregs |= mregs;
		return 0;
	    }
	    if(prm >= 0) {
		mregs |= 1 << prm;	/* mark the parameter as set */
		param[prm].len = s - param[prm].s;
	    }
	    s += mp - mpat;
	    f = mflag;
	    if(!(mflag &= 2)) continue;

	    mflag = 0;
	    {   /* post-process single word match */
		int k = param[prm].len;
		char *w = param[prm].s;
		if(!(f & 1)) {	/* match only if param is single-worded */
		    while(--k >= 0) if(!isalpha(w[k])) {
			dirtyregs |= mregs;
			return 0;
		    }
		} else {	/* unanchored '&' start, take last word */
		    while(k && isalpha(w[k - 1])) k--;
		    param[prm].len -= k;  param[prm].s += k;
		}
	    }
	    continue;
	}
	if(prm >= 0) {		/* '&' at the end of the pattern space */
	    int k;
	    if(mflag & 2) {	/* take first word */
		if(mflag & 1) while(*s && !isalpha(*s)) s++;
		for(k = 0; isalpha(s[k]); k++);
	    } else
		k = strlen(s);
	    mregs |= 1 << prm;
	    param[prm].len = k;  param[prm].s = s;
	}
	break;
    } /* end while */
    if(!(mregs & 1))		/* set &0 to entire line */
	param[0].len = s + strlen(s) - param[0].s;
    {				/* clear dirty params */
	struct param *pp = param;
	int k = dirtyregs & ~mregs;
	dirtyregs = mregs;
	while((k >>= 1))
	    if(pp++, k & 1) pp->len = 0;
    }
    return param;
}

/*
 * execute an action. return 1 if line should be printed,
 * or 0 if it should be suppressed.
 */
static int exec_action(cmd, param)
char *cmd; struct param *param;
{
    if(!strcmp("#gag", cmd)) return 0;
    clear_prompt();
    parse(cmd, 0, param);
    redraw_prompt();
    return 1;
}

/*
 * search for actions to trig on an input line.
 */
int search_action(line, len)
char *line;
int len;
{
    int l = len;
    actionnode *p;
    for(p = actions; p; p = p->next)
	/*
	 * NOTE: we could use mml as an active flag also: toggle actions
	 * on/off with mml = ~mml, and compare (unsigned)mml <= l here.
	 */
	if(p->active && p->mml <= l) {
	    struct param *match = match_action(p->pattern, line);
	    if(match) return exec_action(p->commands, match);
	}
    return 1;
}

/*
 * escape special chars with a backslash
 * return the length of resulting string
 */
int escape_specials(str, buf, spec)
char *str, *buf, *spec;
{
    int k;
    char *s, *d;
    /* NOTE: we dont check for overflow here, it shouldn't happen */
    for(s = str, d = buf; ; *d++ = '\\', *d++ = *s++) {
	if((k = strcspn(s, spec)))
	    do *d++ = *s++; while(--k);
	if(!*s) break;
    }
    *d = 0;
    return d - buf;
}

/*
 * enclose in braces and escape &\; or unbalanced }{ with a backslash,
 * when needed. otherwise just copy the string.
 */
void escape_command(str, buf)
char str[LINELEN], buf[LINELEN];
{
    char *d, *s, *k;
    if(!skip_arg(str, CMDSEP) && skip_arg(str, '\0')) {
	strcpy(buf, str);
	return;
    }
    d = buf; s = k = str; *d++ = '{';
    do {
	if(s == k) {
	    if(*s == CMDSEP || *s == '&' || *s == '\\' || *s == '}' ||
		    (*s == '{' && !(k = skip_arg(s + 1, '}')) && (k = s, 1))) {
		if(d - buf >= LINELEN - 2) break;
		*d++ = '\\';
	    }
	    k++;
	}
	*d++ = *s++;
    } while(d - buf < LINELEN - 1 && (*s || (*d++ = '}', 0)));
    *d = '\0';
}

/*
 * remove escapes (backslashes) from a string
 */
void unescape(s)
char *s;
{
    char *d = s, c;
    do {
	c = *s;
	if(c == '\\') c = *++s;
    } while(s++, *d++ = c);
}

/*
 * make an escape sequence printable (for use in definition-files)
 * NOTE: we should check for buf overflow
 */
void escape_seq(seq, buf)
char *seq, *buf;
{
    char *s, *p;
    unsigned char c;
    /*
     * rules: escape =\^ with \. control chars are written as ^X.
     */
    for(p = buf, s = seq; (c = *s); s++) {
	if(c == '=' || c == '\\' || c == '^')
	    *p++ = '\\';
	else if(c < ' ')
	    *p++ = '^', c |= '@';
#if 0
	else if(!isprint(c)) {
	    sprintf(p, "\\%.3o", (int)c);
	    p += 4;
	    continue;
	}
#endif
	*p++ = c;
    }
    *p = '\0';
}

/*
 * unescape, but also recognize ^c syntax.
 */
void unescape_seq(s)
char *s;
{
    char *d = s, c;
    do switch((c = *s)) {
	/* NOTE: we should skip NUL's here (i.e. ^@) */
	case '^':
	    s[1] = CONTROL(s[1]);
	    c = *++s; break;
	case '\\':
	    c = *++s;
#if 0
	    {
		unsigned int t = ctoi(s[0]);
		if(t < 10) {
		    c = t;
		    t = ctoi(s[1]);
		    if(t < 10) {
			s++; c <<= 3; c += t;
			t = ctoi(s[1]);
			if(t < 10) {
			    s++; c <<= 3; c += t;
			}
		    }
		}
	    }
#endif
    } while(s++, *d++ = c);
}

/*
 * split str into arguments separated by whitespace.
 * parameter 0 contains the whole line.
 */
static void split_arg(str, param)
char *str;
struct param param[10];
{
    int i = 1;
    char *s = str;
    param[0].s = s;
    param[0].len = strlen(s);
    do {
	param[i].s = s = skipspace(s);
	if(!*s) break;
	while(*++s && *s != ' ');
	param[i].len = s - param[i].s;
    } while(++i < 10 && *s++);
    while(i < 10) param[i++].len = 0;
}

/*
 * like strchr, but honour (possibly nested) braces and escaping backslashes.
 * return pointer to the delim in str, or NULL.
 */
static char *skip_arg(str, delim)
char *str;
char delim;
{
    char *s = str - 1;
    for( ; ; ) {
	if(*++s == delim) return s;
	if(!*s || *s == '}') break;
	if(*s != '\\') {
	    if(*s == '{') {
		int nested = 1;
		for( ; ; )	/* inner loop (text within {}) */
		    if(*++s)
			if(*s != '{') {
			    if(*s != '}') {
				if(*s == '\\' && !*++s) return 0;
			    } else if(!--nested) break;
			} else nested++;
		    else return 0;
	    }
	} else if(!*++s) {
	    if(delim) break;
	    return s;
	}
    }
    return 0;
}

/*
 * split the argument into left and right sides, separated by delim
 * return pointer to right hand side, or 0 if no delim was found.
 */
char *chop_arg(str, delim)
char *str, delim;
{
    char *s = skip_arg(str, delim);
    if(s) *s++ = 0;
    return s;
}

/*
 * like chop_arg, but no special handling of {\;
 * used as a replacement of strtok
 */
char *chop_str(str, delim)
char *str, delim;
{
    char *s = str;
    if(*s) while(*s != delim && *++s);
    if(*s) for(*s = 0; *++s == delim; );
    return *s ? s : 0;
}

/*
 * return pointer to next non-blank char
 * FEATURE: only space is considered whitespace.
 * editline is all printable chars anyway.
 */
char *skipspace(p)
char *p;
{
    while(*p == ' ') p++;
    return p;
}

/*
 * return pointer to next non-digit
 */
char *skipnum(p)
char *p;
{
    if(*p == '-') p++;
    while(isdigit(*p)) p++;
    return p;
}
