/*
 * tel.c - telnet protocol communication module for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/telnet.h>
#ifndef TELOPT_NAWS
#define TELOPT_NAWS 31
#endif
#include <arpa/inet.h>
#ifndef NEXT
#include <unistd.h>
#endif

#ifdef TERM
#include "client.h"
#endif /* TERM */

#include "cancan.h"
#include "term.h"

int socket();
int connect();

#ifndef OSF1
unsigned long inet_addr();
#endif

#ifdef TELOPTS
#define TELOPTSTR(n) ((n) > NTELOPTS ? "unknown" : telopts[n])
#endif

static void sendopt();
static void dosubopt();

static int do_naws;		/* negotiate (send our) window size */

#ifndef OMIT_SESSIONS
static int output_socket;	/* to which socket buffer should be sent */
#else
#define output_socket (sockfd)
#endif

/* i wonder, what is the optimum output_buffer size? does it matter? */
static int output_len;		/* number of characters in output_buffer */
static char output_buffer[BUFSIZE];

/*
 * connect to remote host
 * Warning: some voodoo code here
 */
int doconnect(addr, port)
char *addr; int port;
{
    int fd;
#ifndef TERM
    struct sockaddr_in address;
    struct hostent *host_info;

    set_spec_keys(1);	/* we might hang here for a long time... */
    printf("Trying %s... ", addr); fflush(stdout);
#if 0	/* inet_aton() is not quite portable */
    if(inet_aton(addr, &address.sin_addr)) {
#else
    /*
     * NOTE: inet_addr is documented to return -1 for malformed requests,
     * but it is declared to return unsigned long!  Anyway, this works.
     */
    address.sin_addr.s_addr = inet_addr(addr);
    if((int)address.sin_addr.s_addr != -1) {
#endif
	address.sin_family = AF_INET;
    } else if(!(host_info = gethostbyname(addr))) {
	printf("unknown host\n");
	return -1;
    } else {
	memcpy((char*)&address.sin_addr,
		host_info->h_addr, host_info->h_length);
	address.sin_family = host_info->h_addrtype;
    }
    address.sin_port = htons(port);

    if((fd = socket(address.sin_family, SOCK_STREAM, 0)) < 0)
	return syswarn("error creating socket");
    if(fd >= MAXFDSCAN) {
	close(fd);
	printf("not connected (Too many open connections)\n");
	return -1;
    }

    while(connect(fd, (struct sockaddr *)&address, sizeof(address)) < 0)
	if(errno != EINTR) {
	    close(fd);
	    return syswarn("host not responding");
	}
    printf("connected!\n");
    set_spec_keys(0);

#else /* term */

    set_spec_keys(1);	/* we might hang here for a long time... */
    if((fd = connect_server(0)) < 0)
	printf("Unable to connect to term server\n");
    else {
	send_command(fd, C_PORT, 0, "%s:%d", addr, port);
	printf("Connected to term server...\n");
#ifdef TERM_COMPRESS
	send_command(fd, C_COMPRESS, 1, "y");
#endif
	send_command(fd, C_DUMB, 1, 0);
    }
    set_spec_keys(0);
#endif /* term */

    return fd;
}

/*
 * read a maximum of size bytes from remote host
 * using the telnet protocol. return bytes read.
 */
int read_socket(buffer, size)
char *buffer; int size;
{
    /*
     * NOTE: could add a GOTSBTTYPE state to handle the only suboption
     * negotiation we currently support, and do without subopt buffer.
     */
    typedef enum {
	NORMAL = 0, GOTIAC, GOTWILL, GOTWONT, GOTDO, GOTDONT, GOTSB, GOTSBIAC
    } telnetstate;
#ifndef OMIT_SESSIONS
    static telnetstate sock_state[MAXFDSCAN];
    telnetstate state = sock_state[sockfd];
#else
    static telnetstate sock_state;
    telnetstate state = sock_state;
#endif
    static unsigned char subopt[MAXSUBOPT];
    static int subchars;
    int i;
    unsigned char *p, *s;

    if((i = read(sockfd, buffer, size)) <= 0) {
#ifndef OMIT_SESSIONS
	if(i < 0) syswarn("#warning: read from socket");
	sock_state[sockfd] = NORMAL;
	/*
	 * NOTE: we are called from get_remote_output() with both prompt
	 * and input line displayed. And what if we are in LM_CHAR mode??
	 */
	clear_all(); hilite(0);			/* yargh */
	disconnect_sess(sockfd);
	redraw_all();
	return 0;
#else
	if(i < 0) syserr("read from socket");
	exitcancan("#connection closed by remote host.\n");
#endif
    }

    /* scan through the buffer and interpret telnet protocol escapes */
    for(s = p = (unsigned char*)buffer; i; s++, i--) {
	switch(state) {
	default:
	case NORMAL:
	    while(*s == IAC
		    ? (state = GOTIAC, 0)
		    : (*p++ = *s, i > 1)) s++, i--;
	    break;

	case GOTIAC:
	    switch(*s) {
	    case WILL:
		state = GOTWILL; break;
	    case WONT:
		state = GOTWONT; break;
	    case DO:
		state = GOTDO; break;
	    case DONT:
		state = GOTDONT; break;
	    case SB:
		state = GOTSB;
		if(is_main_sess())	/* subopts ignored in subsession */
		    subchars = 0;
		break;
	    case IAC:
		*p++ = IAC;
		state = NORMAL;
		break;
		/*
	    case GA:
		 * NOTE: GA could be handled as end-of-prompt marker one day.
		 * However, prompts have to be identified as such in the
		 * absence of GA anyway, and what is there to gain after all?
		 */
	    default:
		/* ignore the rest of the telnet commands */
#ifdef TELOPTS
		fprintf(stderr, "[skipped IAC <%d>]\n", *s);
#endif
		state = NORMAL;
		break;
	    }
	    break;

	/*
	 * NOTE: perhaps we should handle some more telnet options, like:
	 * - TELOPT_BINARY: but this is used to indicate we should turn off
	 * tty opost, and probably all text processing. The tcp stream is
	 * 8 bit anyway...
	 * - TELOPT_LINEMODE: since cancan is fully operative in linemode
	 * only, we should negotiate the "real" linemode. If that is denied,
	 * set linemode according to klugly SGA option.
	 * - TELOPT_XASCII, TELOPT_RCP: rarely used, why bother?
	 */

	case GOTWILL:
#ifdef TELOPTS
	    fprintf(stderr, "[got WILL %s]\n", TELOPTSTR(*s));
#endif
	    switch(*s) {
	    case TELOPT_ECHO:
		/*
		 * host echoes, turn off echo here
		 * we don't want subsidiary session password entries
		 * to block anything in the main session, though.
		 */
		if(is_main_sess())
		    linemode |= LM_NOECHO;
		sendopt(DO, *s);
		break;
	    case TELOPT_SGA:		/* this can't hurt */
		/*
		 * NOTE: we should send WILL LINEMODE here, and set
		 * LM_CHAR when response is DONT LINEMODE. And if the
		 * response is DO, we need not bother with subsequent
		 * SGA requests at all.
		 */
		if(is_main_sess()) {
		    linemode |= LM_CHAR;
		    set_spec_keys(1);
		}
		sendopt(DO, *s);
		break;
	    default:
		/* don't accept other options */
		sendopt(DONT, *s);
		break;
	    }
	    state = NORMAL;
	    break;

	case GOTWONT:
#ifdef TELOPTS
	    fprintf(stderr, "[got WONT %s]\n", TELOPTSTR(*s));
#endif
	    if(*s == TELOPT_ECHO) {
		/* host no longer echoes, we do it instead */
		if(is_main_sess())
		    linemode &= ~LM_NOECHO;
	    }
	    /* accept any WONT */
	    sendopt(DONT, *s);
	    state = NORMAL;
	    break;

	case GOTDO:
#ifdef TELOPTS
	    fprintf(stderr, "[got DO %s]\n", TELOPTSTR(*s));
#endif
	    if(!is_main_sess()) {
		sendopt(WONT, *s);
	    } else switch(*s) {
	    case TELOPT_SGA:
		linemode |= LM_CHAR;
		set_spec_keys(1);
		/* FALLTHROUGH */
	    case TELOPT_TTYPE:
		sendopt(WILL, *s);
		break;
	    case TELOPT_NAWS:
		sendopt(WILL, *s);
		do_naws = 1;
		send_term_size();
		break;
	    default:
		/* accept nothing else */
		sendopt(WONT, *s);
		break;
	    }
	    state = NORMAL;
	    break;

	case GOTDONT:
#ifdef TELOPTS
	    fprintf(stderr, "[got DONT %s]\n", TELOPTSTR(*s));
#endif
	    if(*s == TELOPT_SGA) {
		if(is_main_sess()) {
		    linemode &= ~LM_CHAR;
		    set_spec_keys(0);
		}
	    }
	    sendopt(WONT, *s);
	    state = NORMAL;
	    break;

	case GOTSB:
	    if(*s == IAC) {
		state = GOTSBIAC;
	    } else if(is_main_sess()) {	/* subopts ignored in subsession */
		if(subchars < MAXSUBOPT)
		    subopt[subchars++] = *s;
	    }
	    break;

	case GOTSBIAC:
	    if(*s == IAC) {
		state = GOTSB;
		if(is_main_sess()) {	/* subopts ignored in subsession */
		    if(subchars < MAXSUBOPT)
			subopt[subchars++] = IAC;
		}
	    } else if(*s == SE) {
		if(is_main_sess()) {	/* subopts ignored in subsession */
		    subopt[subchars] = '\0';
		    dosubopt((char*)subopt);
		}
		state = NORMAL;
	    } else {
		/*
		 * NOTE: telnet would probably try to process the subopt here.
		 * we just ignore it and hope it goes away.
		 */
		to_footnote("#warning: telnet suboption not IAC SE terminated");
		state = NORMAL;
	    }
	    break;
	}
    }
#ifndef OMIT_SESSIONS
    sock_state[sockfd] = state;
#else
    sock_state = state;
#endif
    return (char*)p - buffer;
}

/*
 * send an option negotiation
 * 'what' is one of WILL, WONT, DO, DONT
 */
static void sendopt(what, opt)
int what; unsigned opt;
{
    static unsigned char buf[3] = { IAC, 0, 0 };
    buf[1] = what; buf[2] = opt;
    send_to_host(buf, 3);
#ifdef TELOPTS
    fprintf(stderr, "[sent %s %s]\n", (what == WILL) ? "WILL" :
	   (what == WONT) ? "WONT" :
	   (what == DO) ? "DO" : (what == DONT) ? "DONT" : "error",
	   TELOPTSTR(opt));
#endif
}

/*
 * Send current terminal size (RFC 1073)
 */
void send_term_size()
{
    if(do_naws) {
	static unsigned char buf[9] =
	    { IAC, SB, TELOPT_NAWS, 0, 0, 0, 0, IAC, SE };

	buf[3] = cols >> 8;
	buf[4] = cols & 0xff;
	buf[5] = lines >> 8;
	buf[6] = lines & 0xff;
	send_to_host(buf, 9);
#ifdef TELOPTS
	fprintf(stderr, "[term size %d %d]\n", cols, lines);
#endif
    }
}

/*
 * process suboptions.
 * right now, only terminal type is processed.
 */
static void dosubopt(str)
char *str;
{

    if(str[0] == TELOPT_TTYPE) {
	if(str[1] == 1) {		/* 1 == SEND */
	    char buf[80], *term;
	    /* NOTE: we should record our term type in termcap init */
	    if(!(term = get_env("TERM"))) term = "unknown";
#if defined(_POSIX_SOURCE)
	    send_to_host(buf, sprintf(buf, "%c%c%c%c%.70s%c%c",
		    IAC, SB, TELOPT_TTYPE, 0, term, IAC, SE));	/* 0 == IS */
#else
	    sprintf(buf, "%c%c%c%c%.70s%c%c",
		    IAC, SB, TELOPT_TTYPE, 0, term, IAC, SE);	/* 0 == IS */
	    send_to_host(buf, 4 + strlen(buf + 4));
#endif
#ifdef TELOPTS
	    fprintf(stderr, "[got SB TERMINAL TYPE SEND]\n");
	    fprintf(stderr, "[sent SB TERMINAL TYPE IS %s]\n", term);
#endif
	}
    }
}

/*
 * put data in the output buffer for transmission to the remote host
 */
void send_to_host(data, len)
char *data; int len;
{
    int space;
#ifndef OMIT_SESSIONS
    if(sockfd != output_socket) {	/* is there data to another socket? */
	flush_output_buffer();		/* then flush it */
	output_socket = sockfd;
    }
#endif
    /*
     * BUG: we dont escape IAC's. Meaning we are not 8-bit clean.
     * However, MUME probably has the same bug (furthermore, MUME seems
     * to eat escaped IAC's.) Most notably, this affects the mp messages
     * (MUME local editing protocol). As long as ascii text (or anything
     * except '\377') is being transferred, we are ok...
     */
    while((space = sizeof(output_buffer) - output_len) < len) {
	memcpy(output_buffer + output_len, data, space);
	output_len = sizeof(output_buffer);
	flush_output_buffer();
	data += space;  len -= space;
    }
    memcpy(output_buffer + output_len, data, len);
    output_len += len;
}

/*
 * send all buffered data to the remote host
 */
void flush_output_buffer()
{
    if(output_len) {
	/* NOTE: no guarantee we don't block here indefinitely? */
	while(write(output_socket, output_buffer, output_len) < 0)
	    if(errno != EINTR) { syserr("write to socket"); break; }
	output_len = 0;
    }
}
